"""Core - Implementation details."""

# Built-ins
from typing import Any, Callable, Optional
from datetime import date
from math import inf

# Angets
from ._decorators import loop
from ._helpers import convert_float_to_int, normalize_to_ascii
from ._exceptions import (EmptyStringError, InvalidConfirmationError, InvalidISOFormatError, InvalidIntervalError,
                          OutOfBoundsError, NonFloatingPointError, NonIntegerError)


@loop
def get_non_empty_str(prompt: str = '', warning: Optional[str] = None, **kwargs: Any) -> str:
    """Prompts for a non-empty string.

    :param str prompt: The prompt string.
    :param str warning: User defined warning string. If None, the default warning will be used.
    :param **kwargs kwargs: Keyword arguments for the looping logic.

    :key bool verbose: Warn the user if any exceptions are raised. No warning will be printed unless explicitly set to True.
    :key int attempts: Allowed number of attempts before raising an exception. One by default.

    :return: A non-empty string.

    :raise EmptyStringError: If the input string is empty.
    """
    user_input: str = input(prompt)
    if not user_input.isspace() and len(user_input) != 0:
        return user_input
    else:
        raise EmptyStringError(warning)


@loop
def get_constrained_number(get_number: Callable, within: tuple[float, float], interval: str, prompt: str = '',
                           warning: Optional[str] = None, **kwargs: Any) -> float | int:
    """Prompts for a number within the constraints.

    :param Callable get_number: Function to get the user inputted number (float | int).
    :param tuple within: A tuple representing (lower, upper) in which the input integer must lie within.
    :param str interval: '(' or ')' for non-inclusive, '[' or ']' for inclusive.
    :param str prompt: Prompt for the integer input.
    :param str warning: The warning message if the input floating-point number is out of bounds.

    :key bool verbose: Warn the user if any exceptions are raised. No warning will be printed unless explicitly set to True.
    :key int attempts: Allowed number of attempts before raising an exception. One by default.

    :return: A number within bounds.

    :raise InvalidIntervalError: If the interval value is invalid.
    """
    intervals: list[str] = ['()', '[]', '(]', '[)']
    if interval not in intervals:
        raise InvalidIntervalError(intervals, interval)

    user_input: float | int = get_number(prompt, warning)
    is_within_lower_bound: bool = within[0] < user_input if interval[0] == '(' else within[0] <= user_input
    is_within_upper_bound: bool = within[1] > user_input if interval[1] == ')' else within[1] >= user_input
    if is_within_lower_bound and is_within_upper_bound:
        return user_input
    else:
        raise OutOfBoundsError(warning)


@loop
def get_float(prompt: str = '', warning: Optional[str] = None, **kwargs: Any) -> float:
    """Prompts for a floating-point number.

    :param str prompt: The prompt string.
    :param str warning: User defined warning string. If None, the default warning will be used.
    :param kwargs: Keyword arguments for the looping logic.

    :key bool verbose: Warn the user if any exceptions are raised. No warning will be printed unless explicitly set to True.
    :key int attempts: Allowed number of attempts before raising an exception. One by default.
    """
    try:
        return float(normalize_to_ascii(get_non_empty_str(prompt)))
    except ValueError:
        raise NonFloatingPointError(warning)


@loop
def get_constrained_float(within: tuple[float, float], interval: str, prompt: str = '', warning: Optional[str] = None,
                          **kwargs: Any) -> float:
    """Prompts for a float within the constraints.

    :param tuple within: A tuple representing (lower, upper) in which the input integer must lie within.
    :param str interval: '(' or ')' for non-inclusive, '[' or ']' for inclusive.
    :param str prompt: The prompt string.
    :param str warning: User defined warning string. If None, the default warning will be used.
    :param kwargs: Keyword arguments for the looping logic.

    :key bool verbose: Warn the user if any exceptions are raised. No warning will be printed unless explicitly set to True.
    :key int attempts: Allowed number of attempts before raising an exception. One by default.
    """
    return get_constrained_number(get_float, within, interval, prompt, warning, **kwargs)


@loop
def get_positive_float(prompt: str = '', warning: Optional[str] = None, **kwargs: Any) -> float:
    """Prompts for a positive integer.

    :param str prompt: The prompt string.
    :param str warning: User defined warning string. If None, the default warning will be used.
    :param kwargs: Keyword arguments for the looping logic.

    :key bool verbose: Warn the user if any exceptions are raised. No warning will be printed unless explicitly set to True.
    :key int attempts: Allowed number of attempts before raising an exception. One by default.
    """
    return get_constrained_float((0, inf), '()', prompt, warning, **kwargs)


@loop
def get_non_negative_float(prompt: str = '', warning: Optional[str] = None, **kwargs: Any) -> float:
    """Prompts for a non-negative integer.

    :param str prompt: The prompt string.
    :param str warning: User defined warning string. If None, the default warning will be used.
    :param kwargs: Keyword arguments for the looping logic.

    :key bool verbose: Warn the user if any exceptions are raised. No warning will be printed unless explicitly set to True.
    :key int attempts: Allowed number of attempts before raising an exception. One by default.
    """
    return get_constrained_float((0, inf), '[)', prompt, warning, **kwargs)


@loop
def get_int(prompt: str = '', warning: Optional[str] = None, **kwargs: Any) -> int:
    """Prompts for an integer.

    :param str prompt: The prompt string.
    :param str warning: User defined warning string. If None, the default warning will be used.
    :param kwargs: Keyword arguments for the looping logic.

    :key bool verbose: Warn the user if any exceptions are raised. No warning will be printed unless explicitly set to True.
    :key int attempts: Allowed number of attempts before raising an exception. One by default.
    """
    try:
        return convert_float_to_int(get_float(prompt))
    except ValueError:
        raise NonIntegerError(warning)


@loop
def get_constrained_int(within: tuple[float, float], interval: str, prompt: str = '', warning: Optional[str] = None,
                        **kwargs: Any) -> int:
    """Prompts for an integer within the constraints.

    :param tuple within: A tuple representing (lower, upper) in which the input integer must lie within.
    :param str interval: '(' or ')' for non-inclusive, '[' or ']' for inclusive.
    :param str prompt: The prompt string.
    :param str warning: User defined warning string. If None, the default warning will be used.
    :param kwargs: Keyword arguments for the looping logic.

    :key bool verbose: Warn the user if any exceptions are raised. No warning will be printed unless explicitly set to True.
    :key int attempts: Allowed number of attempts before raising an exception. One by default.
    """
    return get_constrained_number(get_int, within, interval, prompt, warning, **kwargs)


@loop
def get_positive_int(prompt: str = '', warning: Optional[str] = None, **kwargs: Any) -> int:
    """Prompts for a positive integer.

    :param str prompt: The prompt string.
    :param str warning: User defined warning string. If None, the default warning will be used.
    :param kwargs: Keyword arguments for the looping logic.

    :key bool verbose: Warn the user if any exceptions are raised. No warning will be printed unless explicitly set to True.
    :key int attempts: Allowed number of attempts before raising an exception. One by default.
    """
    return get_constrained_int((0, inf), '()', prompt, warning, **kwargs)


@loop
def get_non_negative_int(prompt: str = '', warning: Optional[str] = None, **kwargs: Any) -> int:
    """Prompts for a non-negative integer.

    :param str prompt: The prompt string.
    :param str warning: User defined warning string. If None, the default warning will be used.
    :param kwargs: Keyword arguments for the looping logic.

    :key bool verbose: Warn the user if any exceptions are raised. No warning will be printed unless explicitly set to True.
    :key int attempts: Allowed number of attempts before raising an exception. One by default.
    """
    return get_constrained_int((0, inf), '[)', prompt, warning, **kwargs)


@loop
def get_date(prompt: str = '', warning: Optional[str] = None, **kwargs: Any) -> date:
    """Prompts for a string with valid ISO 8601 formatting.

    :param str prompt: The prompt string.
    :param str warning: User defined warning string. If None, the default warning will be used.
    :param kwargs: Keyword arguments for the looping logic.

    :key bool verbose: Warn the user if any exceptions are raised. No warning will be printed unless explicitly set to True.
    :key int attempts: Allowed number of attempts before raising an exception. One by default.

    :return: A date object.
    """
    try:
        user_input: str = normalize_to_ascii(get_non_empty_str(prompt))
        return date.fromisoformat(user_input)
    except ValueError:
        raise InvalidISOFormatError(warning)


@loop
def get_confirmation(prompt: str = '', warning: Optional[str] = None, selection: Optional[dict[str, bool]] = None,
                     **kwargs: Any) -> bool:
    """Prompts for a valid confirmation has been read.

    :param str prompt: The prompt string.
    :param str warning: User defined warning string. If None, the default warning will be used.
    :param dict selection: A dictionary of options to confirm with. If None, the default confirmation selection options will be used.
    :param kwargs: Keyword arguments for the looping logic.

    :key bool verbose: Warn the user if any exceptions are raised. No warning will be printed unless explicitly set to True.
    :key int attempts: Allowed number of attempts before raising an exception. One by default.

    :return: True if 'yes' or 'y', otherwise False.
    """
    selection = selection or {'yes': True, 'y': True, 'no': False, 'n': False}
    # Make the selection keys case-insensitive.
    selection = {key.lower(): value for key, value in selection.items()}
    try:
        return selection[get_non_empty_str(prompt).strip().lower()]
    except KeyError:
        raise InvalidConfirmationError(warning)



"""Unit tests for Angets."""
# Built-ins
from datetime import date

# Angets
import angets

# Third-party
import pytest

err = angets.exceptions


class TestGetNonEmptyStr:
    def test_exception0(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: '')
        with pytest.raises(err.EmptyStringError):
            angets.get_non_empty_str()

    def test_exception1(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: ' ')
        with pytest.raises(err.EmptyStringError):
            angets.get_non_empty_str()

    def test_exception2(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: 'Test')
        with pytest.raises(err.InvalidAttemptsValueError):
            angets.get_non_empty_str(attempts=0)

    def test_exception3(self, monkeypatch):
        inputs = iter(['', '', ''])
        monkeypatch.setattr('builtins.input', lambda _: next(inputs))
        with pytest.raises(err.AttemptsExceededError):
            angets.get_non_empty_str(attempts=3)

    def test_returned_value0(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: 'Bob')
        result = angets.get_non_empty_str()
        assert result == 'Bob'

    def test_returned_value1(self, monkeypatch):
        inputs = iter(['', 'Java', 'Bob'])
        monkeypatch.setattr('builtins.input', lambda _: next(inputs))
        result = angets.get_non_empty_str(attempts=3)
        assert result == 'Java'


class TestFloat:
    def test_exception0(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: '')
        with pytest.raises(err.NonFloatingPointError):
            angets.get_float()

    def test_exception1(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: 'テスト用')
        with pytest.raises(err.AttemptsExceededError):
            angets.get_float(attempts=100)

    def test_returned_value0(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: '10')
        result = angets.get_float()
        assert result == 10.0

    def test_returned_value1(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: ' 1991.12 ')
        result = angets.get_float()
        assert result == 1991.12

    def test_normalization0(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: '４２０．０２４')
        result = angets.get_float()
        assert result == 420.024

    def test_normalization1(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: 'ー3．1４')
        result = angets.get_float()
        assert result == -3.14


class TestConstrainedFloat:
    def test_exception0(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: '5')
        with pytest.raises(err.OutOfBoundsError):
            angets.get_constrained_float(within=(1, 4), interval='[]')

    def test_exception1(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: '4')
        with pytest.raises(err.OutOfBoundsError):
            angets.get_constrained_float(within=(1, 4), interval='[)')

    def test_exception2(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: '2')
        with pytest.raises(err.OutOfBoundsError):
            angets.get_constrained_float(within=(2, 1), interval='[]')

    def test_exception3(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: '-2.71')
        with pytest.raises(err.OutOfBoundsError):
            angets.get_positive_float()

    def test_exception4(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: '5')
        with pytest.raises(err.InvalidIntervalError):
            angets.get_constrained_float(within=(1, 4), interval='])')

    def test_exception5(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: '5')
        with pytest.raises(err.InvalidIntervalError):
            angets.get_constrained_float(within=(1, 4), interval='{}')

    def test_returned_value0(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: '4')
        result = angets.get_constrained_float(within=(1, 4), interval='[]')
        assert result == 4

    def test_returned_value1(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: '-0.34')
        result = angets.get_constrained_float(within=(-1, 0), interval='()')
        assert result == -0.34

    def test_returned_value2(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: '2.71')
        result = angets.get_positive_float()
        assert result == 2.71

    def test_returned_value3(self, monkeypatch):
        inputs = iter(['Orgil', '3.14', '-6.28'])
        monkeypatch.setattr('builtins.input', lambda _: next(inputs))
        result = angets.get_non_negative_float(attempts=3)
        assert result == 3.14


class TestInt:
    def test_exception0(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: '')
        with pytest.raises(err.NonIntegerError):
            angets.get_int()

    def test_exception1(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: '1.01')
        with pytest.raises(err.NonIntegerError):
            angets.get_int()

    # Huh, I suppose tests are important after all!
    def test_returned_value0(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: '1.0')
        result = angets.get_int()
        assert result == 1

    def test_returned_value1(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: '４２０')
        result = angets.get_int()
        assert result == 420


class TestConstrainedInt:
    def test_exception0(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: '')
        with pytest.raises(err.NonIntegerError):
            angets.get_constrained_int(within=(1, 4), interval='[]')

    def test_exception1(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: '2.5')
        with pytest.raises(err.NonIntegerError):
            angets.get_constrained_int(within=(1, 4), interval='[]')

    def test_exception2(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: '0')
        with pytest.raises(err.OutOfBoundsError):
            angets.get_positive_int()

    def test_returned_value0(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: '1.0')
        result = angets.get_constrained_int(within=(1, 4), interval='[]')
        assert result == 1

    def test_returned_value1(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: '5')
        result = angets.get_positive_int()
        assert result == 5


class TestDate:
    def test_exception0(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: '')
        with pytest.raises(err.InvalidISOFormatError):
            angets.get_date()

    def test_exception1(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: '2024/2/22')
        with pytest.raises(err.InvalidISOFormatError):
            angets.get_date()

    def test_exception2(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: '2024年2月22日')
        with pytest.raises(err.InvalidISOFormatError):
            angets.get_date()

    def test_exception3(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: '   2022/02/22 ')
        with pytest.raises(err.InvalidISOFormatError):
            angets.get_date()

    def test_returned_value0(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: '2024-02-22')
        result = angets.get_date()
        assert result == date(2024, 2, 22)

    def test_returned_value1(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: '２０２４０２２２')
        result = angets.get_date()
        assert result == date(2024, 2, 22)


class TestConfirmation:
    def test_exception0(self, monkeypatch):
        inputs = iter(['', '', ''])
        monkeypatch.setattr('builtins.input', lambda _: next(inputs))
        with pytest.raises(err.AttemptsExceededError):
            angets.get_confirmation(attempts=3)

    # Its little details that come to haunt you...
    def test_exception1(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: 'Q')
        with pytest.raises(err.InvalidConfirmationError):
            angets.get_confirmation()

    def test_return_value0(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: 'y')
        result = angets.get_confirmation()
        assert result == True

    def test_return_value1(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: 'N')
        result = angets.get_confirmation()
        assert result == False

    def test_return_value2(self, monkeypatch):
        monkeypatch.setattr('builtins.input', lambda _: 'Kay')
        result = angets.get_confirmation(selection={'k': True, 'kay': True, 'n': False, 'nay': False})
        assert result == True



"""Decorators for Angets."""

# Built-ins
from typing import Callable
from functools import wraps

# Angets
from ._defaults import ATTEMPTS
from ._helpers import warn
from ._exceptions import AttemptsExceededError, InvalidAttemptsValueError


def loop(function: Callable):
    @wraps(function)
    def wrapper(*args, **kwargs):
        attempts: int = kwargs.get('attempts', ATTEMPTS)
        if attempts <= 0:
            raise InvalidAttemptsValueError(attempts)
        elif attempts == 1:
            return function(*args, **kwargs)
        else:
            for _ in range(attempts):
                try:
                    return function(*args, **kwargs)
                except ValueError as error:
                    if kwargs.get('verbose'):
                        warn(str(error))
                    continue
            else:
                raise AttemptsExceededError(attempts)

    return wrapper



"""Custom exceptions for Angets."""
from typing import Optional


class InvalidAttemptsValueError(ValueError):
    """Invalid number of attempts."""

    def __init__(self, attempts: int) -> None:
        """Create and return a new InvalidAttemptsValueError object."""
        super(InvalidAttemptsValueError, self).__init__(f'{attempts} is not a valid number of attempts.')


class AttemptsExceededError(Exception):
    """Given attempts exceeded."""

    def __init__(self, attempts: int) -> None:
        """Create and return a new AttemptsExceededError object."""
        super(AttemptsExceededError, self).__init__(f'Attempts exceeded, Total attempts: {attempts}')


class InvalidISOFormatError(ValueError):
    """Invalid ISO format."""

    def __init__(self, warning: Optional[str]) -> None:
        """Create and return a new InvalidISOFormatError object."""
        if warning is None:
            warning = 'Invalid ISO format. Example: (1970-01-01)'

        super(InvalidISOFormatError, self).__init__(warning)


class InvalidConfirmationError(ValueError):
    """Invalid confirmation string."""

    def __init__(self, warning: Optional[str]) -> None:
        """Create and return a new InvalidConfirmationError object."""
        if warning is None:
            warning = 'Invalid confirmation string.'

        super(InvalidConfirmationError, self).__init__(warning)


class InvalidIntervalError(Exception):
    """Invalid interval."""

    def __init__(self, valid_intervals: list[str], invalid_interval: str) -> None:
        """Create and return a new InvalidIntervalError object."""
        super(InvalidIntervalError, self).__init__(
            f'Invalid interval: {invalid_interval}\nValid intervals: {" or ".join(valid_intervals)}')


class OutOfBoundsError(ValueError):
    """Value not within bounds."""

    def __init__(self, warning: Optional[str]) -> None:
        """Create and return a new OutOfBoundsError object."""
        if warning is None:
            warning = 'Value not within bounds.'

        super(OutOfBoundsError, self).__init__(warning)


class EmptyStringError(ValueError):
    """Empty string."""

    def __init__(self, warning: Optional[str]) -> None:
        """Create and return a new EmptyStringError object."""
        if warning is None:
            warning = 'Input is empty. Please input a valid string.'

        super(EmptyStringError, self).__init__(warning)


class NonFloatingPointError(ValueError):
    """Non-floating-point number."""

    def __init__(self, warning: Optional[str]) -> None:
        """Create and return a new NonFloatingPointError object."""
        if warning is None:
            warning = 'Not a floating-point number. Please input a valid floating-point number.'

        super(NonFloatingPointError, self).__init__(warning)


class NonIntegerError(ValueError):
    """Non-integer number."""

    def __init__(self, warning: Optional[str]) -> None:
        """Create and return a new NonIntegerError object."""
        if warning is None:
            warning = 'Not an integer. Please input a valid integer number.'

        super(NonIntegerError, self).__init__(warning)



"""Angets (Ankha's Gets): Functions for user input."""
import importlib.metadata

__version__: str = importlib.metadata.version('angets')

from angets._core import (
    get_non_empty_str, get_constrained_number, get_float, get_constrained_float, get_positive_float,
    get_non_negative_float, get_int, get_constrained_int, get_positive_int, get_non_negative_int, get_confirmation,
    get_date
)
import angets._decorators as decorators
import angets._helpers as helpers
import angets._exceptions as exceptions



"""Utility functions for Angets."""

# Built-ins
from unicodedata import normalize
from re import sub

# Angets
from ._exceptions import NonIntegerError


def warn(warning: str) -> None:
    """Prints the warning message the stream, if there is one."""
    if warning:
        print(warning)


def convert_float_to_int(float_number: float, warning: str | None = None) -> int:
    """Return a float as an integer if said float can be converted into an integer without loss of data.

    :raise NonIntegerError: If float is not an integer.
    """
    if float_number.is_integer():
        return int(float_number)
    else:
        raise NonIntegerError(warning)


def normalize_to_ascii(non_ascii_string: str) -> str:
    """Convert a Japanese full-width number to half-width."""
    return sub('[ー－―—‐]', '-', normalize('NFKC', non_ascii_string))
